#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Token {
    Semi,
    Equals,
    Comma,
    Plus,
    Minus,
    Ampersand,
    Star,
    Number(String),
    Identifier(String),
    Keyword(Keyword),
    OpenParanthesis,
    CloseParanthesis,
    OpenSquiglyBracket,
    CloseSquiglyBracket,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Keyword {
    Return,
    Function,
    Let,
    Assign,
    If,
}

enum TokenFinishedUnfinished {
    Unfinished(Option<Token>),
    Finished(Token),
}

use std::collections::HashMap;

use self::TokenFinishedUnfinished::*;

fn generate_keyword_map() -> HashMap<String, Keyword> {
    let mut keywords: HashMap<String, Keyword> = HashMap::new();

    use Keyword as K;
    keywords.insert("return".to_string(), K::Return);
    keywords.insert("fn".to_string(), K::Function);
    keywords.insert("let".to_string(), K::Let);
    keywords.insert("assign".to_string(), K::Assign);
    keywords.insert("if".to_string(), K::If);

    keywords
}

fn tokenise_current_character(
    current_character: char,
    current_token_type: Option<Token>,
) -> TokenFinishedUnfinished {
    let keywords = generate_keyword_map();

    match current_token_type {
        Some(Token::Number(ref v)) => {
            let mut current_characters = v.clone();

            if current_character.is_digit(10) {
                current_characters.push(current_character);
                return Unfinished(Some(Token::Number(current_characters)));
            }

            Finished(Token::Number(current_characters))
        }
        Some(Token::Identifier(ref v)) => {
            let mut current_characters = v.clone();

            if current_character.is_alphanumeric() {
                current_characters.push(current_character);
                return Unfinished(Some(Token::Identifier(current_characters)));
            }

            match keywords.get(&current_characters) {
                Some(keyword) => Finished(Token::Keyword(keyword.clone())),
                None => Finished(Token::Identifier(current_characters)),
            }
        }
        None => {
            if current_character.is_digit(10) {
                return Unfinished(Some(Token::Number(String::from(current_character))));
            }

            if current_character.is_alphabetic() {
                return Unfinished(Some(Token::Identifier(String::from(current_character))));
            }

            {
                use Option::Some as s;
                use Token as T;
                use TokenFinishedUnfinished::Unfinished as uf;

                uf(match current_character {
                    ';' => s(T::Semi),
                    '(' => s(T::OpenParanthesis),
                    ')' => s(T::CloseParanthesis),
                    '{' => s(T::OpenSquiglyBracket),
                    '}' => s(T::CloseSquiglyBracket),
                    '=' => s(T::Equals),
                    ',' => s(T::Comma),
                    '+' => s(T::Plus),
                    '-' => s(T::Minus),
                    '&' => s(T::Ampersand),
                    '*' => s(T::Star),
                    _ => None,
                })
            }
        }
        Some(v) => Finished(v),
    }
}

pub fn tokenise(input: String) -> Vec<Token> {
    let mut tokens: Vec<Token> = vec![];

    let mut current_character: char;
    let mut current_token_type: Option<Token> = None;

    let mut i: usize = 0;
    while i < input.len() {
        current_character = input.as_bytes()[i] as char;

        match tokenise_current_character(current_character, current_token_type.clone()) {
            Unfinished(v) => {
                current_token_type = v.clone();
            }
            Finished(v) => {
                tokens.push(v);
                current_token_type = None;
                i -= 1;
            }
        }

        i += 1;
    }

    tokens
}
